#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "meminfo.h"

static GtkWidget *main_canvas;
static pid_t target_pid;
static int need_wait_child = 0;
static guint timerID;
static int dirty_mark = 0;

#define MAIN_CANVAS_WIDTH 650
#define MAIN_CANVAS_HEIGHT 800

#define MAREA_X_OFFSET 20
#define MAREA_Y_OFFSET 20
#define MAREA_WIDTH 256
#define MAREA_HEIGHT 760

#define LABEL_X_OFFSET 350
#define LABEL_Y_OFFSET 20

#define UPDATE_INTERVAL 300

static int update_interval = UPDATE_INTERVAL;

struct vma_area_map {
  struct memvmainfo_struct vma;
  int disp_offs;
  int disp_size;
  int npages;
  int npages_allocated;
  struct mempageinfo_struct *pages;
  struct vma_area_map *next;
};

static struct vma_area_map *area_cache = NULL;
static struct vma_area_map *area_map_list = NULL;

static void usage(void)
{
    fprintf(stderr, "Usage: gmeminfo [options] -pid process-id\n");
    fprintf(stderr, "       gmeminfo [options] command args...\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -dirty		Display dirty pages\n");
    fprintf(stderr, "  -interval msec	Specify update interval\n");
    fprintf(stderr, "  -help			Display this message\n");
    exit(1);
}

static void gmem_exit(void)
{
  int status;
  if(need_wait_child) {
    status = 0;
    wait(&status);
    exit(status);
  }
  exit(0);
}

static struct vma_area_map *new_area_map(unsigned long baseaddr)
{
  struct vma_area_map *area;
  struct vma_area_map *prev;

  prev = NULL;
  area = area_cache;
  while(area != NULL) {
    if(area->vma.vm_start == baseaddr)
      break;
    prev = area;
    area = area->next;
  }

  if(area == NULL) {
    area = (struct vma_area_map *)malloc(sizeof(struct vma_area_map));
    area->npages = 0;
    area->npages_allocated = 0;
    area->pages = NULL;
  }
  else {
    if(prev == NULL)
      area_cache = area_cache->next;
    else
      prev->next = area->next;
  }
  return area;
}

static void load_area_map(void)
{
  static struct vma_area_map *head, *tail;
  struct memvmainfo_struct vma_buffer[1];
  struct memvmainfo target;
  struct mempageinfo pageinfo;
  int fd;
  int pagesize = getpagesize();
  int count;

  fd = open(MEMINFOPATH, O_RDONLY);
  if(fd < 0) {
    perror(MEMINFOPATH);
    gmem_exit();
  }

  area_cache = area_map_list;
  area_map_list = NULL;
  head = tail = NULL;
  target.pid = target_pid;
  target.offset = 0;
  target.vma = vma_buffer;
  pageinfo.pid = target_pid;
  count = 0;
  while(1) {
    unsigned long vm_size;
    target.nvmas = 1;
    if(ioctl(fd, GETMEMVMAINFO, &target) == -1) {
      close(fd);
      gmem_exit();
    }
    if(!target.nvmas) {
      break;
    }
    vm_size = vma_buffer[0].vm_end - vma_buffer[0].vm_start;
#if 0
        printf("%08lx-%08lx: %08lx: (%ld) %s\n",
	   vma_buffer[0].vm_start,
	   vma_buffer[0].vm_end,
	   vma_buffer[0].vm_flags,
	   vm_size,
	   vma_buffer[0].filename);
#endif

    pageinfo.npages = vm_size / pagesize;

    if(head == NULL ||
       strcmp(tail->vma.filename, vma_buffer[0].filename) != 0 ||
	tail->vma.vm_end != vma_buffer[0].vm_start) {
      if(head == NULL)
	head = tail = new_area_map(vma_buffer[0].vm_start);
      else
	tail = tail->next = new_area_map(vma_buffer[0].vm_start);
      tail->next = NULL;
      memcpy(&tail->vma, vma_buffer, sizeof(vma_buffer[0]));
      if(tail->npages_allocated < pageinfo.npages) {
	tail->pages = (struct mempageinfo_struct *)realloc(tail->pages,
							   pageinfo.npages * sizeof(struct mempageinfo_struct));
	tail->npages_allocated = pageinfo.npages;
      }
      tail->npages = pageinfo.npages;
      pageinfo.addr = tail->vma.vm_start;
      pageinfo.page = tail->pages;
      if(ioctl(fd, GETMEMPAGEINFO, &pageinfo) == -1) {
	close(fd);
	gmem_exit();
      }
      if(pageinfo.npages != tail->npages) {
	fprintf(stderr, "Warnning: number of pages missmatch: %08lx-%08lx (%ld): %d,%d\n",
		tail->vma.vm_start, tail->vma.vm_end, vm_size,
		pageinfo.npages, tail->npages);
      }
    } else {
      int total_npages;
      int save_npages = pageinfo.npages;

      total_npages = pageinfo.npages + tail->npages;
      if(tail->npages_allocated < total_npages) {
	tail->pages = (struct mempageinfo_struct *)realloc(tail->pages,
							   total_npages * sizeof(struct mempageinfo_struct));
	tail->npages_allocated = total_npages;
      }
      pageinfo.addr = vma_buffer[0].vm_start;
      pageinfo.page = tail->pages + tail->npages;

      if(ioctl(fd, GETMEMPAGEINFO, &pageinfo) == -1) {
	close(fd);
	gmem_exit();
      }

      if(pageinfo.npages != save_npages) {
	fprintf(stderr, "Warnning: number of pages missmatch: %08lx-%08lx (%ld): %d,%d\n",
		vma_buffer[0].vm_start, vma_buffer[0].vm_end, vm_size,
		pageinfo.npages, save_npages);
      }

      tail->npages += pageinfo.npages;
      tail->vma.vm_end = vma_buffer[0].vm_end;
      tail->vma.vm_flags |= vma_buffer[0].vm_flags;
    }


#if 0
    if(head == NULL)
      head = tail = new_area_map(vma_buffer[0].vm_start);
    else
      tail = tail->next = new_area_map(vma_buffer[0].vm_start);
    tail->next = NULL;
    memcpy(&tail->vma, vma_buffer, sizeof(vma_buffer[0]));
    if(tail->npages < pageinfo.npages) {
      tail->pages = (struct mempageinfo_struct *)realloc(tail->pages,
							pageinfo.npages * sizeof(struct mempageinfo_struct));
    }
    tail->npages = pageinfo.npages;
    pageinfo.addr = tail->vma.vm_start;
    pageinfo.page = tail->pages;
    if(ioctl(fd, GETMEMPAGEINFO, &pageinfo) == -1) {
      close(fd);
      gmem_exit();
    }
    if(pageinfo.npages != tail->npages) {
      fprintf(stderr, "Warnning: number of pages missmatch: %08lx-%08lx (%ld): %d,%d\n",
	      tail->vma.vm_start, tail->vma.vm_end, vm_size,
	      pageinfo.npages, tail->npages);
    }
#endif

    target.offset++;
    count++;
  }
  close(fd);

  if(!count)
    gmem_exit();

  area_map_list = head;
  while(area_cache) {
    struct vma_area_map *next = area_cache->next;
    free(area_cache->pages);
    free(area_cache);
    area_cache = next;
  }
}

static int update_area_map(void)
{
  int count;
  static struct vma_area_map *ptr;
  unsigned long mem_total, memsum;

  load_area_map();
  count = 0;
  mem_total = 0;
  for(ptr = area_map_list; ptr; ptr = ptr->next) {
    count++;
    mem_total += ptr->vma.vm_end - ptr->vma.vm_start;
  }

  memsum = 0;
  for(ptr = area_map_list; ptr; ptr = ptr->next) {
    ptr->disp_offs = (int)((double)MAREA_HEIGHT * memsum / mem_total);
    memsum += ptr->vma.vm_end - ptr->vma.vm_start;
  }

  for(ptr = area_map_list; ptr->next; ptr = ptr->next) {
    ptr->disp_size = ptr->next->disp_offs - ptr->disp_offs;
  }
  ptr->disp_size = MAREA_HEIGHT - ptr->disp_offs;
  return count;
}

#define COLOR_RING_SIZE 4
gint update_callback( gpointer data )
{
  GdkWindow *win = main_canvas->window;
  GtkStyle *style = main_canvas->style;

  static GdkFont *font;
  static GdkGC *gc = NULL;
  static struct vma_area_map *area;
  double label_pos, label_incr;
  static GdkPixmap *pm;
  static GdkColor color_ring[COLOR_RING_SIZE];
  static GdkColor color_ring2[COLOR_RING_SIZE];
  char buff[BUFSIZ];
  int i, area_index;

  i = update_area_map();
  if(i == 0)
    return FALSE;
  label_incr = (double)MAREA_HEIGHT / (i - 1);
  if(label_incr < 14.0)
    label_incr = 14.0;

  if(gc == NULL) {
    GdkColormap* colormap;

    /* font = style->font; */
    font = gdk_font_load("fixed");

    gc = gdk_gc_new(win);
    gdk_gc_set_font(gc, font);
    pm = gdk_pixmap_new(win, MAIN_CANVAS_WIDTH, MAIN_CANVAS_HEIGHT, -1);

    gdk_color_parse("#8fef8f", color_ring + 0);
    gdk_color_parse("#8fafef", color_ring + 1);
    gdk_color_parse("#efafaf", color_ring + 2);
    gdk_color_parse("#efef5f", color_ring + 3);

    gdk_color_parse("#4f7f4f", color_ring2 + 0);
    gdk_color_parse("#4f5f7f", color_ring2 + 1);
    gdk_color_parse("#7f5f5f", color_ring2 + 2);
    gdk_color_parse("#7f7f30", color_ring2 + 3);

    colormap = gdk_colormap_get_system();

    for(i = 0; i < COLOR_RING_SIZE; i++)
      gdk_colormap_alloc_color(colormap, color_ring + i, FALSE, TRUE);
    for(i = 0; i < COLOR_RING_SIZE; i++)
      gdk_colormap_alloc_color(colormap, color_ring2 + i, FALSE, TRUE);
  }

  gdk_draw_rectangle(pm, style->fg_gc[GTK_STATE_NORMAL],
		     TRUE, 0, 0, MAIN_CANVAS_WIDTH, MAIN_CANVAS_HEIGHT);

  /*    gtk_draw_box(style, pm,
	       GTK_STATE_NORMAL, GTK_SHADOW_ETCHED_IN,
		MAREA_X_OFFSET - 0, MAREA_Y_OFFSET - 0,
		MAREA_WIDTH + 1,
		MAREA_HEIGHT + 1); */

  label_pos = (double)MAREA_Y_OFFSET;
  area_index = 0;
  for(area = area_map_list; area; area = area->next) {
    int y = MAREA_Y_OFFSET + area->disp_offs;
    char *label = "[no map]";

    gdk_gc_set_foreground(gc, &color_ring[area_index % COLOR_RING_SIZE]);

    /*
    gdk_draw_line(pm, style->fg_gc[GTK_STATE_NORMAL],
		  MAREA_X_OFFSET, y,
		  MAREA_X_OFFSET + MAREA_WIDTH, y);
    */

    gdk_draw_line(pm, gc,
		  MAREA_X_OFFSET + MAREA_WIDTH + 3, y,
		  LABEL_X_OFFSET - 2, (int)label_pos - 4);

    if(*area->vma.filename)
      label = area->vma.filename;
    else if(area->vma.vm_flags & 0x00000100)
      label = "[stack]";
    sprintf(buff, "%08lx-%08lx %s",
	    area->vma.vm_start, area->vma.vm_end, label);
    gdk_draw_string(pm, font, gc, LABEL_X_OFFSET, (int)label_pos, buff);
    if(area->disp_size >= 1) {
      static char pagemark[MAREA_HEIGHT];
      int size = area->disp_size;
      int np = area->npages;
      int i, j, last_idx = 0, last_flag = 0;
      memset(pagemark, 0, size);
      for(i = 0; i < np; i++) {
	int idx = size * i / np;
	char flag;

	if(area->pages[i].u.flags.present) {
	  flag = 1;
	  if(dirty_mark && area->pages[i].u.flags.dirty) {
	    flag |= 2;
	  }
	} else
	  flag = 0;
	j = last_idx + 1;
	while(j < idx) {
	  pagemark[j++] |= last_flag;
	}
	pagemark[idx] |= flag;
	last_flag = flag;
	last_idx = idx;
      }

      j = last_idx + 1;
      while(j < size) {
	pagemark[j++] |= last_flag;
      }

      for(i = 0; i < size; i++) {
	if(pagemark[i]) {
	  gdk_gc_set_foreground(gc, &color_ring[area_index % COLOR_RING_SIZE]);
	  gdk_draw_line(pm, gc,
			MAREA_X_OFFSET, y + 1 + i,
			MAREA_X_OFFSET + MAREA_WIDTH, y + 1 + i);
	  if(pagemark[i] & 2) {
	    gdk_gc_set_foreground(gc, &color_ring2[area_index % COLOR_RING_SIZE]);
	    gdk_draw_line(pm, gc,
			  MAREA_X_OFFSET, y + 1 + i,
			  MAREA_X_OFFSET + MAREA_WIDTH/2, y + 1 + i);
	  }
	}
      }
    }
    label_pos += label_incr;
    area_index++;
  }

  gdk_draw_pixmap(win, gc, pm, 0, 0, 0, 0,
		  MAIN_CANVAS_WIDTH, MAIN_CANVAS_HEIGHT);

  gtk_timeout_remove(timerID);
  timerID = gtk_timeout_add(update_interval, update_callback, NULL);

  return TRUE;
}

static gint main_canvas_event(GtkWidget	*widget, 
			      GdkEvent	*event,
			      gpointer	data)
{
  if(event->type == GDK_EXPOSE && event->expose.count == 0) {
    gtk_timeout_remove(timerID);
    timerID = gtk_timeout_add(1, update_callback, NULL);
  }
  return TRUE;
}

static int start_gmeminfo()
{
  gtk_set_locale();

  /* create a new window */
  main_canvas = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(main_canvas, 
		       MAIN_CANVAS_WIDTH,
		       MAIN_CANVAS_HEIGHT);
  gtk_signal_connect(GTK_OBJECT(main_canvas), "expose_event",
		     GTK_SIGNAL_FUNC(main_canvas_event),
		     NULL);


  /* and the window */
  gtk_widget_show(main_canvas);

  timerID = gtk_timeout_add(update_interval, update_callback, NULL);

  /* All GTK applications must have a gtk_main(). Control ends here
   * and waits for an event to occur (like a key press or
   * mouse event). */

  gtk_main();
  
  return(0);
}

int main(int argc, char *argv[])
{
  gtk_init(&argc, &argv);

  argc--;
  argv++;

  while(argc > 0) {
    if(strcmp(argv[0], "-help") == 0)
      usage();
    if(strcmp(argv[0], "-dirty") == 0) {
      dirty_mark = 1;
    } else if(strcmp(argv[0], "-interval") == 0) {
      argc--;
      argv++;
      if(!argc)
	usage();
      update_interval = atoi(argv[0]);
    } else
      break;
    argc--;
    argv++;
  }

  if(!argc) {
    usage();
  }

  if(strcmp(argv[0], "-pid") == 0) {
    if(argc <= 1)
      usage();
    target_pid = atoi(argv[1]);
    start_gmeminfo();
  } else {
    target_pid = fork();
    if(target_pid != 0) {
      /* parent */
      need_wait_child = 1;
      sleep(1);
      start_gmeminfo();
      gmem_exit();
    } else {
      if(execvp(argv[0], argv) == -1) {
	/* child */
	perror(argv[0]);
	kill(getppid(), SIGTERM);
	exit(1);
      }
    }
  }
  return 0;
}
