#include <linux/module.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/pgtable.h>
#include <linux/utsname.h>
#include <asm/uaccess.h>
#include <linux/string.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include "meminfo.h"

#ifdef DEBUG
#define PRINTK(x) printk x
#else
#define PRINTK(x)
#endif

static inline _syscall3(int,mknod,const char *,filename,int,mode,dev_t,dev)
static inline _syscall1(int,unlink,const char *,filename)

static int major = 0;
MODULE_PARM(major,"i");

static void (* mmput_func)(struct mm_struct *) = NULL;

static struct mm_struct *meminfo_pid2mm(pid_t pid)
{
  struct task_struct *task;
  struct mm_struct *mm;

  if(pid == 0 || pid == current->pid)
    task = current;
  else {
    if(!suser())
      return (struct mm_struct *)ERR_PTR(-EACCES);
    if((task = find_task_by_pid(pid)) == NULL)
      return (struct mm_struct *)ERR_PTR(-ESRCH);
  }

  task_lock(task);
  if(task->state == TASK_ZOMBIE) {
    task_unlock(task);
    return (struct mm_struct *)ERR_PTR(-ESRCH);
  }

  mm = task->mm;
  if(mm)
    atomic_inc(&mm->mm_users);
  task_unlock(task);
  return task->mm;
}

static int meminfo_pte_range(pmd_t *pmd,
			     unsigned long offset,
			     unsigned long address,
			     unsigned long size,
			     struct mempageinfo_struct *page,
			     int count)
{
  pte_t * pte;
  unsigned long end;
  unsigned long offs;
  int n = 0;

  if(pmd == NULL || pmd_none(*pmd) || pmd_bad(*pmd))
    pte = NULL;
  else
    pte = pte_offset(pmd, address);
  offs = (__pmd_offset(address) << PMD_SHIFT) + offset;
  address &= ~PMD_MASK;
  end = address + size;
  if (end > PMD_SIZE)
    end = PMD_SIZE;

  if(pte == NULL) {
    do {
      page->u.allflags = 0;
      page++;
      n++;
      address += PAGE_SIZE;
    } while (address < end && n < count);
  } else {
    do {
      pte_t pe = *pte++;

      page->u.allflags = 0;
      if(!pte_none(pe)) {
	page->u.flags.available = 1;
	if(pte_present(pe)) {
	  page->u.flags.present = 1;
	  if(pte_dirty(pe)) {
	    page->u.flags.dirty = 1;
	  }
	}
      }
      page++;
      n++;
      address += PAGE_SIZE;
    } while (address < end && n < count);
  }
  return n;
}


static int meminfo_pmd_range(pgd_t *pgd,
			     unsigned long address,
			     unsigned long size,
			     struct mempageinfo_struct *page,
			     int count)
{
  pmd_t * pmd;
  unsigned long end;
  unsigned long offs;
  int n = 0;

  if(pgd_none(*pgd) || pgd_bad(*pgd))
    pmd = NULL;
  else
    pmd = pmd_offset(pgd, address);
  offs = __pgd_offset(address) << PGDIR_SHIFT;
  address &= ~PGDIR_MASK;
  end = address + size;
  if(end > PGDIR_SIZE)
    end = PGDIR_SIZE;
  do {
    n += meminfo_pte_range(pmd, offs, address, end - address,
			   page + n, count - n);
    address = (address + PMD_SIZE) & PMD_MASK;
    if(pmd)
      pmd++;
  } while(address < end && n < count);
  return n;
}

static int meminfo_pgd_range(pgd_t *pgd,
			     unsigned long address,
			     unsigned long end,
			     struct mempageinfo_struct *page,
			     int count)
{
  int n = 0;
  while(address < end && n < count) {
    n += meminfo_pmd_range(pgd, address, end - address,
			   page + n, count - n);
    address = (address + PGDIR_SIZE) & PGDIR_MASK;
    pgd++;
  }
  return n;
}

static unsigned long lookup_System_map(const char *symbol_name)
{
  char buff[1024];
  int fd = -1;
  mm_segment_t fs;
  unsigned long addr = 0;
  int n, symlen;

  sprintf(buff,
	  "/boot/System.map-%s", system_utsname.release);

  fs = get_fs();
  set_fs(get_ds());

  if((fd = open(buff, 0, 0)) < 0) {
    printk(KERN_ERR "Can not open %s\n", buff);
    goto finish;
  }

  symlen = strlen(symbol_name);
  while((n = read(fd, buff, sizeof(buff)-1)) > 0) {
    int backoff = 0;
    char *symptr;

    while(n > 0 && buff[n - 1] != '\n') {
      n--;
      backoff++;
    }
    if(n > 0)
      buff[n - 1] = '\0';
    else
      goto finish;
    if(backoff)
      lseek(fd, 1, -backoff);
    if((symptr = strstr(buff, symbol_name)) != NULL &&
       (symptr[symlen] == '\n' || symptr[symlen] == '\0')) {
      unsigned long x = 0;
      symptr -= 4;
      while(buff <= symptr && *symptr != '\n')
	symptr--;
      symptr++;
      while(1) {
	int c = *symptr++;
	if('0' <= c && c <= '9')
	  c = c - '0';
	else if('a' <= c && c <= 'f')
	  c = c - 'a' + 10;
	else
	  break;
	x = ((x << 4) | c);
      }
      addr = x;
      break;
    }
  }

 finish:
  if(fd >= 0)
    close(fd);
  set_fs(fs);

  return addr;
}

static int mempageinfo_ioctl(struct mempageinfo *param)
{
  struct mempageinfo info;
  struct mm_struct *mm;
  unsigned long addr;
  int count, err;
  struct vm_area_struct * vma;
  pgd_t *pgd;

  if((err = verify_area(VERIFY_READ, param, sizeof(*param))) < 0)
    return err;
  if((err = verify_area(VERIFY_WRITE, param, sizeof(*param))) < 0)
    return err;
  info = *param;
  if((err = verify_area(VERIFY_WRITE, info.page,
			info.npages * sizeof(struct mempageinfo_struct))) < 0)
    return err;
  mm = meminfo_pid2mm(info.pid);
  if(!mm) {
    info.npages = 0;
    *param = info;
    return 0;
  }
  if(IS_ERR(mm))
    return PTR_ERR(mm);

  down(&mm->mmap_sem);
  addr = (info.addr & PAGE_MASK);
  count = info.npages;
  if((vma = find_vma(mm, addr)) == NULL) {
    up(&mm->mmap_sem);
    mmput_func(mm);
    info.npages = 0;
    *param = info;
    return 0;
  }

  pgd = pgd_offset(mm, vma->vm_start);
  info.npages = meminfo_pgd_range(pgd, addr, vma->vm_end,
				  info.page, count);
  up(&mm->mmap_sem);
  mmput_func(mm);
  *param = info;
  return 0;
}

static int memvmainfo_ioctl(struct memvmainfo *param)
{
  struct memvmainfo info;
  struct mm_struct *mm;
  struct vm_area_struct *vma;
  struct memvmainfo_struct *p;
  int i, err, offset;

  if((err = verify_area(VERIFY_READ, param, sizeof(*param))) < 0)
    return err;
  if((err = verify_area(VERIFY_WRITE, param, sizeof(*param))) < 0)
    return err;
  info = *param;
  if((err = verify_area(VERIFY_WRITE, info.vma,
			info.nvmas * sizeof(struct memvmainfo_struct))) < 0)
    return err;

  mm = meminfo_pid2mm(info.pid);
  if(!mm) {
    info.nvmas = 0;
    *param = info;
    return 0;
  }
  if(IS_ERR(mm))
    return PTR_ERR(mm);

  down(&mm->mmap_sem);

  offset = info.offset;
  vma = mm->mmap;
  while(offset > 0 && vma) {
    offset--;
    vma = vma->vm_next;
  }
  p = info.vma;
  for(i = 0;
      i < info.nvmas && vma;
      i++, vma = vma->vm_next) {
    p[i].vm_start = vma->vm_start;
    p[i].vm_end = vma->vm_end;
    p[i].vm_flags = vma->vm_flags;
    if(vma->vm_file == NULL)
      p[i].filename[0] = '\0';
    else {
      struct qstr *fn = &vma->vm_file->f_dentry->d_name;
      int len;
      len = fn->len;
      if(len > sizeof(p[i].filename) - 1)
	len = sizeof(p[i].filename) - 1;
      memcpy(p[i].filename, fn->name, len);
      p[i].filename[len] = '\0';
    }
  }
  up(&mm->mmap_sem);
  mmput_func(mm);
  info.nvmas = i;
  *param = info;
  return 0;
}

static int meminfo_ioctl(struct inode *inode, struct file *filp,
			 unsigned int cmd,
			 unsigned long param)
{


  switch(cmd) {
  case GETMEMPAGEINFO:
    return mempageinfo_ioctl((struct mempageinfo *)param);
  case GETMEMVMAINFO:
    return memvmainfo_ioctl((struct memvmainfo *)param);
  }
  return -EINVAL;
}

static struct file_operations meminfo_fops = {
  ioctl: meminfo_ioctl,
};

void cleanup_module(void);
int init_module(void)
{
  mm_segment_t fs;
  long err;

  mmput_func = (void (*)(struct mm_struct *))lookup_System_map("mmput");
  if(mmput_func == NULL) {
    printk(KERN_ERR "System map is not fount\n");
    return -EBUSY;
  }

  major = register_chrdev(0, "meminfo", &meminfo_fops);
  if(major < 0)
    return major;

  fs = get_fs();
  set_fs(get_ds());
  err = mknod(MEMINFOPATH, S_IFCHR | 0400, MKDEV(major,0));
  set_fs(fs);

  if(err < 0) {
    printk(KERN_ERR "Can not mknod: errno=%ld\n", -err);
    cleanup_module();
    return err;
  }

  return 0;
}


void cleanup_module(void)
{
  mm_segment_t fs;

  fs = get_fs();
  set_fs(get_ds());
  unlink(MEMINFOPATH);
  set_fs(fs);

  unregister_chrdev(major, "meminfo");
}
